/*
 * Copyright 2011 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2011-2022 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "constify.h"

static struct {
	const char *writable_section;
	const char *readonly_section;
	const char *readonly_bss_section;
} constified_sections[] = {
	{".init.data",			".init.rodata",			".init.rodata"},
	{".ref.data",			".ref.rodata",			".ref.rodata"},
	{".meminit.data",		".meminit.rodata",		".meminit.rodata"},
	{".data..read_mostly",		RODATA_MUTABLE_READ_MOSTLY,	BSS_MUTABLE},
	{".data..ro_after_init",	RODATA_MUTABLE, 		BSS_MUTABLE},
};

static bool is_constified_var(varpool_node_ptr node)
{
	const_tree var = NODE_DECL(node);
	const_tree type = get_inner_type(var);
	const char *name;

	if (node->alias)
		return false;

	if (TREE_READONLY(var))
		return false;

	if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var)))
		return true;

	switch (TREE_CODE(type)) {
	default:
		return false;

	case RECORD_TYPE:
	case UNION_TYPE:
		if (DECL_EXTERNAL(var))
			return false;
		break;
	}

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type)))
		return true;

	name = get_decl_section_name(var);
	if (name && is_const_section(name))
		return false;

	if (!TYPE_READONLY(type) || !C_TYPE_FIELDS_READONLY(type))
		return false;

	return TYPE_CONSTIFY_VISITED(type);
}

static void handle_section_mismatch(varpool_node_ptr node)
{
	tree var = NODE_DECL(node);
	location_t loc = DECL_SOURCE_LOCATION(var);
	const_tree type = get_inner_type(var);
	const_tree section;
	const char *old_section_name, *new_section_name;
	size_t i;

	old_section_name = get_decl_section_name(var);
	section = lookup_attribute("section", DECL_ATTRIBUTES(var));

	// check that the two ways of storing section info are consistently set/unset
	if (!section != !old_section_name) {
		fprintf(stderr, "DECL_SECTION: [%s] != section attr: [%s]\n", old_section_name ? old_section_name : "", section ? TREE_STRING_POINTER(section) : "");
		dump_varpool_node(stderr, node);
		debug_tree(var);
		gcc_unreachable();
	}

	// the implicit default section is .data or .bss
	if (!old_section_name) {
		new_section_name = DECL_INITIAL(var) ? RODATA_MUTABLE : BSS_MUTABLE;

		if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type)))
			return;
		if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
			//if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var)))
			//	error_at(loc, "constify: variable %qD with constified mutable type %qT has mutable_const attribute", var, type);
			if (verbose)
				inform(loc, "constify: variable %qD with constified mutable type %qT moved into section %qs", var, type, new_section_name);
			set_decl_section_name(var, new_section_name);
		} else if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var))) {
			if (verbose)
				inform(loc, "constify: constified mutable variable %qD moved into section %qs", var, new_section_name);
			set_decl_section_name(var, new_section_name);
		} else if (POINTER_TYPE_P(type) && lookup_attribute("do_const", TYPE_ATTRIBUTES(TREE_TYPE(type))) != NULL_TREE) {
			if (verbose)
				inform(loc, "constify: pointer to constified variable %qD moved into section %qs", var, new_section_name);
			set_decl_section_name(var, new_section_name);
		} else if (POINTER_TYPE_P(type) && lookup_attribute("mutable_const", TYPE_ATTRIBUTES(TREE_TYPE(type))) != NULL_TREE) {
			if (verbose)
				inform(loc, "constify: pointer to mutable variable %qD moved into section %qs", var, new_section_name);
			set_decl_section_name(var, new_section_name);
		} else if (!TYPE_CONSTIFY_VISITED(type)) {
			dump_varpool_node(stderr, node);
			debug_tree(var);
			debug_tree(type);
			gcc_unreachable();
		}
		return;
	}

//	gcc_assert(!TREE_CHAIN(section));
	gcc_assert(TREE_VALUE(section));

	section = TREE_VALUE(TREE_VALUE(section));
	gcc_assert(!strcmp(TREE_STRING_POINTER(section), old_section_name));

	if (is_const_section(old_section_name))
		return;

	for (i = 0; i < ARRAY_SIZE(constified_sections); i++) {
		if (strcmp(constified_sections[i].writable_section, old_section_name))
			continue;

		new_section_name = DECL_INITIAL(var) ? constified_sections[i].readonly_section : constified_sections[i].readonly_bss_section;
		if (verbose)
			inform(loc, "constify: constified mutable variable %qD moved into section %s", var, new_section_name);
		set_decl_section_name(var, new_section_name);
		return;
	}

	dump_varpool_node(stderr, node);
	debug_tree(var);
	debug_tree(var);
	fflush(stderr);
	error_at(loc, "constified variable %qD placed into writable section %s", var, old_section_name);
}

/*
 * this works around a gcc bug/feature where uninitialized globals
 * are moved into the .bss section regardless of any constification
 * see gcc/varasm.c:bss_initializer_p()
 */
static void fix_initializer(tree var)
{
	tree type = TREE_TYPE(var);

	if (DECL_INITIAL(var))
		return;

	DECL_INITIAL(var) = build_constructor(type, NULL);
}

void handle_global_variables(void *event_data __unused, void *data __unused)
{
	varpool_node_ptr node;

	FOR_EACH_VARIABLE(node) {
		if (!is_constified_var(node))
			continue;

		handle_section_mismatch(node);
		fix_initializer(NODE_DECL(node));
	}
}
