/*
 * Copyright 2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * various out-of-line functions from gcc-common.h
 */

#include "utilities.h"

static void __add_type_attr(tree type, const char *attr, tree args)
{
	tree oldattr;

	if (type == NULL_TREE)
		return;
	oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
	if (oldattr != NULL_TREE) {
		gcc_assert(TREE_VALUE(oldattr) == args || TREE_VALUE(TREE_VALUE(oldattr)) == TREE_VALUE(args));
		return;
	}

	TYPE_ATTRIBUTES(type) = copy_list(TYPE_ATTRIBUTES(type));
	TYPE_ATTRIBUTES(type) = tree_cons(get_identifier(attr), args, TYPE_ATTRIBUTES(type));
}

__visible void add_type_attr(tree type, const char *attr, tree args)
{
	tree main_variant = TYPE_MAIN_VARIANT(type);

	__add_type_attr(TYPE_CANONICAL(type), attr, args);
	__add_type_attr(TYPE_CANONICAL(main_variant), attr, args);
	__add_type_attr(main_variant, attr, args);

	for (type = TYPE_NEXT_VARIANT(main_variant); type; type = TYPE_NEXT_VARIANT(type)) {
		tree oldattr;

		oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
		if (oldattr != NULL_TREE)
			gcc_assert(TREE_VALUE(oldattr) == args || TREE_VALUE(TREE_VALUE(oldattr)) == TREE_VALUE(args));
		else
			TYPE_ATTRIBUTES(type) = TYPE_ATTRIBUTES(main_variant);

		__add_type_attr(TYPE_CANONICAL(type), attr, args);
	}
}

static void __update_type_attr(tree type, const char *attr, tree args)
{
	tree oldattr;

	if (type == NULL_TREE)
		return;
	oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
	if (oldattr == NULL_TREE)
		return;

	TYPE_ATTRIBUTES(type) = copy_list(TYPE_ATTRIBUTES(type));
	oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
	gcc_assert(oldattr);
	TREE_VALUE(oldattr) = args;
}

__visible void update_type_attr(tree type, const char *attr, tree args)
{
	tree main_variant = TYPE_MAIN_VARIANT(type);

	__update_type_attr(TYPE_CANONICAL(type), attr, args);
	__update_type_attr(TYPE_CANONICAL(main_variant), attr, args);
	__update_type_attr(main_variant, attr, args);

	for (type = TYPE_NEXT_VARIANT(main_variant); type; type = TYPE_NEXT_VARIANT(type)) {
		tree oldattr;

		oldattr = lookup_attribute(attr, TYPE_ATTRIBUTES(type));
		if (oldattr != NULL_TREE)
			gcc_assert(TREE_VALUE(oldattr) == args || TREE_VALUE(TREE_VALUE(oldattr)) == TREE_VALUE(args));
		else
			TYPE_ATTRIBUTES(type) = TYPE_ATTRIBUTES(main_variant);

		__add_type_attr(TYPE_CANONICAL(type), attr, args);
	}
}

__visible basic_block plugin_split_edge(edge e)
{
	basic_block bb;

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));

	bb = split_edge(e);
	gcc_assert(single_succ_p(e->dest));

	if (!dom_info_available_p(CDI_POST_DOMINATORS))
		return bb;

	// split_edge doesn't adjust CDI_POST_DOMINATORS but we need this information
	set_immediate_dominator(CDI_POST_DOMINATORS, bb, single_succ(bb));

	if (get_immediate_dominator(CDI_POST_DOMINATORS, single_pred(bb)) == single_succ(bb)) {
		edge_iterator ei;
		edge f;

		FOR_EACH_EDGE(f, ei, single_pred(bb)->succs) {
			if (f == single_pred_edge(bb))
				continue;

			if (!dominated_by_p(CDI_POST_DOMINATORS, f->dest, single_pred(bb)))
				break;
		}

		if (!f)
			set_immediate_dominator(CDI_POST_DOMINATORS, single_pred(bb), bb);
	}

	return bb;
}

__visible basic_block nearest_common_dominator_for_sbitmap(enum cdi_direction dir, const_sbitmap blocks)
{
	unsigned int i;
	sbitmap_iterator sbi;
	basic_block dom;

	if (-1 == sbitmap_first_set_bit(blocks))
		return NULL;

	dom = BASIC_BLOCK_FOR_FN(cfun, sbitmap_first_set_bit(blocks));
	EXECUTE_IF_SET_IN_SBITMAP(blocks, 0, i, sbi)
		if (dom != BASIC_BLOCK_FOR_FN(cfun, i))
			dom = nearest_common_dominator(dir, dom, BASIC_BLOCK_FOR_FN(cfun, i));

	return dom;
}

__visible tree build_const_char_string(int len, const char *str)
{
	tree cstr, elem, index, type;

	cstr = build_string(len, str);
	elem = build_type_variant(char_type_node, 1, 0);
	index = build_index_type(size_int(len - 1));
	type = build_array_type(elem, index);
	build_pointer_type(type);
	TREE_TYPE(cstr) = type;
	TREE_CONSTANT(cstr) = 1;
	TREE_READONLY(cstr) = 1;
	TREE_STATIC(cstr) = 1;
	return cstr;
}

__visible location_t gimple_inline_point_location(const_gimple_stmt_ptr g)
{
	location_t loc = UNKNOWN_LOCATION;
	tree block;

	block = gimple_block(g);
	gcc_assert(block == NULL_TREE || TREE_CODE(block) == BLOCK);

	for (; block != NULL_TREE && TREE_CODE(block) == BLOCK; block = BLOCK_SUPERCONTEXT(block)) {
		tree ao = BLOCK_ABSTRACT_ORIGIN(block);

		if (ao == NULL_TREE)
			continue;
		if (ao == current_function_decl)
			break;

		switch (TREE_CODE(ao)) {
		case BLOCK:
			if (BLOCK_SOURCE_LOCATION(block) != UNKNOWN_LOCATION)
				loc = BLOCK_SOURCE_LOCATION(block);
			continue;

		case FUNCTION_DECL:
			loc = BLOCK_SOURCE_LOCATION(block);
			continue;

		default:
			debug_gimple_stmt(g);
			debug_tree(block);
			debug_tree(ao);
			gcc_unreachable();
		}
	}
	gcc_assert(gimple_block(g) == NULL_TREE || block != NULL_TREE);

	if (loc == UNKNOWN_LOCATION)
		loc = gimple_expansion_point_location(g);
	if (loc == UNKNOWN_LOCATION)
		loc = gimple_location(g);
	return loc;
}

__visible bool insn_dominates_insn_p(enum cdi_direction dir, rtx_insn *insn1, rtx_insn *insn2)
{
	basic_block bb1, bb2;

	if (insn1 == insn2)
		return true;

	bb1 = BLOCK_FOR_INSN(insn1);
	bb2 = BLOCK_FOR_INSN(insn2);

	if (bb1 == bb2) {
		rtx_insn *insn;

		FOR_BB_INSNS(bb1, insn) {
			if (insn == insn2)
				return dir == CDI_POST_DOMINATORS;
			if (insn == insn1)
				return dir == CDI_DOMINATORS;
		}

		gcc_unreachable();
	}

	return dominated_by_p(dir, bb2, bb1);
}

// create the equivalent of
// asm volatile("" : : : "memory");
// or
// asm("" : "+rm"(var));
// or
// asm("" : "+rm"(var) : : "memory");
__visible gimple_stmt_ptr gimple_barrier(tree var, bool full)
{
	gimple_stmt_ptr stmt;
	vec<tree, va_gc> *inputs = NULL;
	vec<tree, va_gc> *outputs = NULL;
	vec<tree, va_gc> *clobbers = NULL;

	if (var) {
		tree input;
		tree output;

		input = build_tree_list(NULL_TREE, build_const_char_string(2, "0"));
		input = chainon(NULL_TREE, build_tree_list(input, var));
		VEC_safe_push(tree, gc, inputs, input);

		output = build_tree_list(NULL_TREE, build_const_char_string(4, "=rm"));
		if (!VAR_P(var)) {
			gcc_assert(SSA_NAME_VAR(var));
			var = make_ssa_name(SSA_NAME_VAR(var), NULL);
		}
		output = chainon(NULL_TREE, build_tree_list(output, var));
		VEC_safe_push(tree, gc, outputs, output);
	}

	if (full) {
		tree clobber;

		clobber = build_tree_list(NULL_TREE, build_const_char_string(7, "memory"));
		VEC_safe_push(tree, gc, clobbers, clobber);
	}

	stmt = gimple_build_asm_vec("", inputs, outputs, clobbers, NULL);
	if (full && !var)
		gimple_asm_set_volatile(as_a_gasm(stmt), true);
	if (var && !VAR_P(var))
		SSA_NAME_DEF_STMT(var) = stmt;
	return stmt;
}

__visible bool tree_for_decl_and_aliases(tree decl, bool (*callback)(tree, void *), void *data)
{
	cgraph_node_ptr node;
	unsigned int i;
	alias_pair *p;

	if (callback(decl, data))
		return true;

	if (!alias_pairs)
		return false;

	node = cgraph_get_node(decl);
	if (!node)
		return false;

	FOR_EACH_VEC_SAFE_ELT(alias_pairs, i, p) {
		if (TREE_CODE(p->decl) != FUNCTION_DECL)
			continue;
		if (node != cgraph_node_for_asm(p->target))
			continue;

		if (tree_for_decl_and_aliases(p->decl, callback, data))
			return true;
	}

	return false;
}

__visible basic_block get_single_pred_entry_block(void)
{
	basic_block entry_bb;

	entry_bb = ENTRY_BLOCK_PTR_FOR_FN(cfun);
	gcc_assert(single_succ_p(entry_bb));
	entry_bb = single_succ(entry_bb);
	if (single_pred_p(entry_bb))
		return entry_bb;

//	gcc_assert(bb_loop_depth(entry_bb) || (entry_bb->flags & BB_IRREDUCIBLE_LOOP));
	gcc_assert(bb_any_loop(entry_bb));
	entry_bb = plugin_split_edge(single_succ_edge(entry_bb));
	gcc_assert(single_pred_p(entry_bb));
	gcc_assert(single_succ_p(entry_bb));
	gcc_assert(!bb_any_loop(entry_bb));
	return entry_bb;
}

__visible bool has_trait(tree decl, const char *trait)
{
	tree attr;

	attr = lookup_attribute("traits", DECL_ATTRIBUTES(decl));
	if (attr != NULL_TREE)
		attr = TREE_VALUE(attr);

	for (; attr; attr = TREE_CHAIN(attr)) {
		if (strcmp(trait, TREE_STRING_POINTER(TREE_VALUE(attr))))
			continue;

		return true;
	}

	return false;
}

__visible void clone_late_inlined_nodes(cgraph_edge_p edge)
{
	cgraph_edge_p next, callee_edge;

	gcc_assert(gimple_has_body_p(NODE_DECL(edge->callee)));
	gcc_assert(!DECL_FUNCTION_PERSONALITY(NODE_DECL(edge->callee)));
	gcc_assert(opt_for_fn(NODE_DECL(edge->callee), flag_strict_aliasing) == opt_for_fn(NODE_DECL(edge->caller), flag_strict_aliasing));
	gcc_assert(!NODE_GLOBAL(edge->callee)->inlined_to);

	edge->inline_failed = CIF_OK;
	clone_inlined_nodes(edge, true, false, NULL);

	for (callee_edge = edge->callee->callees; callee_edge; callee_edge = next) {
		cgraph_node_ptr callee = callee_edge->callee->ultimate_alias_target();

		next = callee_edge->next_callee;
		if (!lookup_attribute("late_inline", DECL_ATTRIBUTES(NODE_DECL(callee))))
			continue;
		clone_late_inlined_nodes(callee_edge);
	}
}
