/*
 * Copyright 2021 by Open Source Security, Inc.
 * Copyright 2021-2024 by PaX Team <pageexec@freemail.hu>
 *
 * Licensed under the GPL v2
 *
 * Implement the 'leaf_only' function attribute.
 *
 * This attribute -- in contrast to the already existing 'leaf' one -- is more
 * than just a hint to the compiler. It actually enforces that a function is a
 * leaf function and errors out if not.
 */

#include "utilities.h"

static tree handle_leaf_only_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl = *node;

	*no_add_attrs = true;
	if (TREE_CODE(decl) != FUNCTION_DECL) {
		location_t loc = DECL_SOURCE_LOCATION(decl);
		warning_at(loc, OPT_Wattributes, "%qE attribute only applies to functions", name);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec leaf_only_attr = ATTRIBUTE_SPEC_INIT("leaf_only", 0, 0, true, false, false, handle_leaf_only_attribute, false, NULL);

static unsigned int leaf_only_execute(void)
{
	rtx_insn *insn;

	if (!lookup_attribute("leaf_only", DECL_ATTRIBUTES(current_function_decl)))
		return 0;

	// This test is moot, as it ignores tail calls...
//	if (crtl->is_leaf)	// aka 'current_function_is_leaf'
//		return 0;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		location_t loc;
		tree call_decl;
		rtx body;

		if (INSN_DELETED_P(insn))
			continue;

		// rtl match (call_insn:TI 96 611 505 4 (call (mem:QI (symbol_ref:DI ("ftrace_likely_update") [flags 0x41] <function_decl 0x7f36393c0100 ftrace_likely_update>) [0 ftrace_likely_update S1 A8])
		if (!CALL_P(insn))
			continue;

		loc = INSN_LOCATION(insn);

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);
		if (GET_CODE(body) == SET)
			body = SET_SRC(body);
		gcc_assert(GET_CODE(body) == CALL);

		body = XEXP(body, 0);
		gcc_assert(MEM_P(body));

		body = XEXP(body, 0);
		call_decl = NULL_TREE;
		if (GET_CODE(body) == SYMBOL_REF) {
			gcc_assert(SYMBOL_REF_FUNCTION_P(body));
			call_decl = SYMBOL_REF_DECL(body);
		}

		if (call_decl == NULL_TREE)
			error_at(loc, "leaf_only function %qD contains unknown call", current_function_decl);
		else /* if (!TREE_THIS_VOLATILE(call_decl)) */
			error_at(loc, "leaf_only function %qD calls %qD", current_function_decl, call_decl);
	}

	return 0;
}

#define PASS_NAME leaf_only
#define NO_GATE
#include "gcc-generate-rtl-pass.h"
