/*
 * Copyright 2020-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * implement the 'naked' function attribute for gcc 6-7
 */

#include "utilities.h"

#if defined(TARGET_386) && BUILDING_GCC_VERSION >= 6000 && BUILDING_GCC_VERSION < 8000
static tree handle_naked_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl;
	location_t loc;

	*no_add_attrs = true;
	decl = *node;
	loc = DECL_SOURCE_LOCATION(decl);

	if (TREE_CODE(decl) != FUNCTION_DECL) {
		warning_at(loc, OPT_Wattributes, "%qE attribute only applies to functions", name);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec naked_attr = ATTRIBUTE_SPEC_INIT("naked", 0, 0, true, false, false, handle_naked_attribute, false, NULL);

static bool (*old_allocate_stack_slots_for_args)(void);
static bool (*old_warn_func_return)(tree);
static rtx_insn *(*old_gen_sibcall_epilogue)(void);
static rtx_insn *(*old_gen_epilogue)(void);
static rtx_insn *(*old_gen_prologue)(void);
static bool (*old_have_return)(void);

static bool utilities_allocate_stack_slots_for_args(void)
{
	return lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl)) ? false : old_allocate_stack_slots_for_args();
}

static bool utilities_warn_func_return(tree decl)
{
	return lookup_attribute("naked", DECL_ATTRIBUTES(decl)) ? false : old_warn_func_return(decl);
}

static rtx_insn *utilities_gen_sibcall_epilogue(void)
{
	rtx_insn *val;

	if (!lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl)))
		return old_gen_sibcall_epilogue();

	start_sequence();
	emit_insn(gen_trap());
	val = get_insns();
	end_sequence();
	return val;
}

static rtx_insn *utilities_gen_epilogue(void)
{
	rtx_insn *val;

	if (!lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl)))
		return old_gen_epilogue();

	start_sequence();
	emit_insn(gen_trap());
	val = get_insns();
	end_sequence();
	return val;
}

static rtx_insn *utilities_gen_prologue(void)
{
	return lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl)) ? NULL : old_gen_prologue();
}

static bool utilities_have_return(void)
{
	return lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl)) ? false : old_have_return();
}

void naked_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	old_allocate_stack_slots_for_args = targetm.calls.allocate_stack_slots_for_args;
	targetm.calls.allocate_stack_slots_for_args = utilities_allocate_stack_slots_for_args;

	old_warn_func_return = targetm.warn_func_return;
	targetm.warn_func_return = utilities_warn_func_return;

	old_gen_sibcall_epilogue = targetm.gen_sibcall_epilogue;
	targetm.gen_sibcall_epilogue = utilities_gen_sibcall_epilogue;

	old_gen_epilogue = targetm.gen_epilogue;
	targetm.gen_epilogue = utilities_gen_epilogue;

	old_gen_prologue = targetm.gen_prologue;
	targetm.gen_prologue = utilities_gen_prologue;

	old_have_return = targetm.have_return;
	targetm.have_return = utilities_have_return;
}
#endif
