/*
 * Copyright 2021-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * implement the 'no_stack_protector' function attribute for gcc 5-10
 *
 * as a sideeffect, it also fixes a potential issue on mips/r6000/xtensa
 * where FRAME_GROWS_DOWNWARD depends on the global SSP flag but does not
 * take the function attr into account (it's still wrong for SANITIZE_ADDRESS)
 */

#include "utilities.h"
#include "options.h"

#if BUILDING_GCC_VERSION < 11000
static tree handle_no_stack_protector_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl;
	location_t loc;

	*no_add_attrs = true;
	decl = *node;
	loc = DECL_SOURCE_LOCATION(decl);

	if (TREE_CODE(decl) != FUNCTION_DECL) {
		warning_at(loc, OPT_Wattributes, "%qE attribute only applies to functions", name);
		return NULL_TREE;
	}

	if (lookup_attribute("stack_protect", DECL_ATTRIBUTES(decl))) {
		error_at(loc, "%qE attribute is incompatible with the 'stack_protect' attribute", name);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec no_stack_protector_attr = ATTRIBUTE_SPEC_INIT("no_stack_protector", 0, 0, true, false, false, handle_no_stack_protector_attribute, false, NULL);

#endif

static __typeof__(flag_stack_protect) flag_stack_protect_orig;

void no_stack_protector_start(void *event_data, void *data __unused)
{
	gcc_assert(current_function_decl);

	flag_stack_protect_orig = flag_stack_protect;
	if (!lookup_attribute("no_stack_protector", DECL_ATTRIBUTES(current_function_decl)))
		return;

	if (lookup_attribute("stack_protect", DECL_ATTRIBUTES(current_function_decl))) {
		error_at(DECL_SOURCE_LOCATION(current_function_decl), "'no_stack_protector' attribute is incompatible with the 'stack_protect' attribute");
		return;
	}

	flag_stack_protect = 0;
}

void no_stack_protector_end(void *event_data, void *data __unused)
{
	flag_stack_protect = flag_stack_protect_orig;
}
