/*
 * Copyright 2012-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to provide various features for older gcc versions or other plugins
 * <gcc-8:   emergency dump on ICE
 * plugins:  context tracking infrastructure
 *           symbolic kernel constants via ksyms.def
 *
 */

#include "utilities.h"

__visible int plugin_is_GPL_compatible;

static struct plugin_info utilities_plugin_info = {
	.version	= "202401101500",
	.help		= "asm_include=<file>\t\tinclude <file> into all translation units to provide various asm macros\n"
			  "align-init-text-bits=<bits>\t\talign .init functions to 2^<bits>\n"
			  "report-null-deref\t\treport potential NULL pointer dereferences left around\n"
			  "convert-nolocals\t\tmove nolocal variables off the stack\n"
			  "report-nolocals\t\treport nolocal variable conversions\n"
#ifdef TARGET_386
			  "sls-regparm\t\tmodified regparm=3 ABI not making use of EAX\n"
#endif
};

static void register_attributes(void *event_data, void *data)
{
	register_attribute(&context_attr);
	register_attribute(&diagnose_attr);
	register_attribute(&late_inline_attr);
	register_attribute(&leaf_only_attr);
	register_attribute(&maybe_used_attr);
#if defined(TARGET_386) && BUILDING_GCC_VERSION >= 6000 && BUILDING_GCC_VERSION < 8000
	register_attribute(&naked_attr);
#endif
#if defined(TARGET_386) && BUILDING_GCC_VERSION < 7000
	if (TARGET_64BIT)
		register_attribute(&no_caller_saved_registers_attr);
#endif
	register_attribute(&noderef_attr);
	register_attribute(&nolocal_attr);
#if BUILDING_GCC_VERSION < 8000
	register_attribute(&no_sanitize_attr);
#endif
#if BUILDING_GCC_VERSION < 12000
	register_attribute(&no_sanitize_coverage_attr);
#endif
#if BUILDING_GCC_VERSION < 11000
	register_attribute(&no_stack_protector_attr);
#endif
	register_attribute(&read_once_attr);
	register_attribute(&write_once_attr);
	register_attribute(&section_type_attr);
	register_attribute(&traits_attr);
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	int i;
	bool enable_context_tracking = true;
	bool convert_nolocals = false;
#ifdef TARGET_386
	bool sls_regparm = false;
#endif

//	PASS_INFO(context, "ssa", 1, PASS_POS_INSERT_AFTER);
//	PASS_INFO(context, "phiprop", 1, PASS_POS_INSERT_AFTER);
#if BUILDING_GCC_VERSION >= 12000
	PASS_INFO(context, "early_uninit", 1, PASS_POS_INSERT_AFTER);
#else
	PASS_INFO(context, "*early_warn_uninitialized", 1, PASS_POS_INSERT_AFTER);
#endif

	PASS_INFO(diagnose, "*warn_function_noreturn", 1, PASS_POS_INSERT_AFTER);

#if BUILDING_GCC_VERSION < 8000
	PASS_INFO(fix_esra, "esra", 1, PASS_POS_INSERT_BEFORE);
#endif

#ifdef TARGET_386
	PASS_INFO(fix_tailc, "barriers", 1, PASS_POS_INSERT_AFTER);
#endif

#if BUILDING_GCC_VERSION < 13000
	PASS_INFO(late_inline, "cdtor", 1, PASS_POS_INSERT_AFTER);
#endif

	PASS_INFO(leaf_only, "final", 1, PASS_POS_INSERT_BEFORE);

#if BUILDING_GCC_VERSION < 8000
	PASS_INFO(no_sanitize_before_einline, "einline", 0, PASS_POS_INSERT_BEFORE);
	PASS_INFO(no_sanitize_after_einline,  "einline", 0, PASS_POS_INSERT_AFTER);
	PASS_INFO(no_sanitize_before_inline,   "inline", 0, PASS_POS_INSERT_BEFORE);
	PASS_INFO(no_sanitize_after_inline,    "inline", 0, PASS_POS_INSERT_AFTER);
#endif

	PASS_INFO(once_before, "dse", 0, PASS_POS_INSERT_BEFORE);
	PASS_INFO(once_after, "dse", 0, PASS_POS_INSERT_AFTER);

	PASS_INFO(unmark_maybe_used_functions, "visibility", 1, PASS_POS_INSERT_AFTER);

	PASS_INFO(early_discover_nolocal_variables, "einline", 1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(ipa_discover_nolocal_variables, "inline", 1, PASS_POS_INSERT_AFTER);
	PASS_INFO(convert_nolocal_variables, "fnsplit", 1, PASS_POS_INSERT_BEFORE);

	PASS_INFO(dead_copy_elimination, "ssa", 1, PASS_POS_INSERT_AFTER);

#if defined(TARGET_386) && BUILDING_GCC_VERSION < 6000
	PASS_INFO(no_caller_saved_registers_gen_call, "vregs", 1, PASS_POS_INSERT_BEFORE);
#endif

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &utilities_plugin_info);

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "nocontext")) {
			enable_context_tracking = false;
			continue;
		}
		if (!strcmp(argv[i].key, "asm_include")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			asm_include = xstrdup(argv[i].value);
			continue;
		}
		if (!strcmp(argv[i].key, "align-init-text-bits")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			align_init_text_bits = atoi(argv[i].value);
			if (align_init_text_bits == 0 || align_init_text_bits > 8) {
				error(G_("invalid value supplied for option '-fplugin-arg-%s-%s' (%s)"), plugin_name, argv[i].key, align_init_text_bits ? "too big" : "too small");
				continue;
			}
			align_init_text_bits = (1U << align_init_text_bits) * BITS_PER_UNIT;
			continue;
		}
		if (!strcmp(argv[i].key, "report-null-deref")) {
			report_null_deref = true;
			continue;
		}
		if (!strcmp(argv[i].key, "convert-nolocals")) {
			convert_nolocals = true;
			continue;
		}
		if (!strcmp(argv[i].key, "report-nolocals")) {
			report_nolocals = true;
			continue;
		}
#ifdef TARGET_386
		if (!strcmp(argv[i].key, "sls-regparm")) {
			if (TARGET_64BIT) {
				error(G_("'-fplugin-arg-%s-%s' requires 32 bit target"), plugin_name, argv[i].key);
				continue;
			}
			sls_regparm = true;
			continue;
		}
#endif
		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

#if BUILDING_GCC_VERSION < 8000
	old_internal_error = global_dc->internal_error;
	global_dc->internal_error = emergency_dump_internal_error_function;
#endif

	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);
	if (enable_context_tracking)
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &context_pass_info);
	if (asm_include)
		register_callback(plugin_name, PLUGIN_START_UNIT, asm_include_start_unit, NULL);

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &diagnose_pass_info);

#if BUILDING_GCC_VERSION < 8000
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &fix_esra_pass_info);
#endif

#ifdef TARGET_386
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &fix_tailc_pass_info);
#endif

	late_inline_register_hook();

#if BUILDING_GCC_VERSION >= 13000
	register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, late_inline_delay_inlines, NULL);
#else
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &late_inline_pass_info);
#endif

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &leaf_only_pass_info);

#if BUILDING_GCC_VERSION < 8000
	if (flag_sanitize & SANITIZE_ADDRESS) {
		no_sanitize_register_hook();
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &no_sanitize_before_einline_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &no_sanitize_after_einline_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &no_sanitize_before_inline_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &no_sanitize_after_inline_pass_info);
	}
#endif

	register_callback(plugin_name, PLUGIN_ALL_PASSES_START, no_stack_protector_start, NULL);
	register_callback(plugin_name, PLUGIN_ALL_PASSES_END, no_stack_protector_end, NULL);

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &once_before_pass_info);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &once_after_pass_info);

	register_callback(plugin_name, PLUGIN_FINISH_DECL, ksyms_find_decls, NULL);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_ksyms);

	if (align_init_text_bits)
		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, align_init_text_sections, NULL);

	register_callback(plugin_name, PLUGIN_START_UNIT, section_type_start_unit, NULL);

#if defined(TARGET_386) && BUILDING_GCC_VERSION >= 6000 && BUILDING_GCC_VERSION < 8000
	register_callback(plugin_name, PLUGIN_START_UNIT, naked_start_unit, NULL);
#endif

#if defined(TARGET_386) && BUILDING_GCC_VERSION < 7000
	if (TARGET_64BIT) {
		register_callback(plugin_name, PLUGIN_START_UNIT, no_caller_saved_registers_start_unit, NULL);
#if BUILDING_GCC_VERSION < 6000
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &no_caller_saved_registers_gen_call_pass_info);
#endif
	}
#endif

#ifdef TARGET_386
	if (sls_regparm)
		register_callback(plugin_name, PLUGIN_START_UNIT, regparm_start_unit, NULL);
#endif

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &unmark_maybe_used_functions_pass_info);

	if (convert_nolocals) {
		register_callback(plugin_name, PLUGIN_START_UNIT, nolocal_start_unit, NULL);
		register_callback(plugin_name, PLUGIN_FINISH_TYPE, nolocal_finish_type, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &early_discover_nolocal_variables_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &ipa_discover_nolocal_variables_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &convert_nolocal_variables_pass_info);
	}
	register_callback(plugin_name, PLUGIN_START_UNIT, nolocal_typemap_allocate, NULL);
	register_callback(plugin_name, PLUGIN_ALL_PASSES_START, nolocal_typemap_cleanup, NULL);

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &dead_copy_elimination_pass_info);

	return 0;
}
